/*--------------------------------------------------------------------
 * 
 *  Driver main file
 *  This file contains all the driver code.
 *
 *--------------------------------------------------------------------*/

#ifndef __KERNEL__
# define __KERNEL__
#endif

#include <linux/config.h>  /* for CONFIG_PCI */

#ifdef MODULE
# include <linux/module.h>
# include <linux/version.h>
#else
# define MOD_INC_USE_COUNT
# define MOD_DEC_USE_COUNT
#endif

#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/slab.h>
//#include <linux/malloc.h>
#include <linux/pci.h>
#include <linux/poll.h>
#include <linux/ioport.h>
#include <linux/stat.h>
#include <linux/delay.h>
#include <asm/io.h>    /* for inb(), outb() etc. */
#include <asm/system.h>
#include <asm/pgtable.h>
#include <linux/bigphysarea.h>

#include "aa.h"
#include "amcc.h"
//#include "allocator.h"

#define BUFFER_SIZE (1<<10) //in kilobytes
#define P2A_SIZE	(1<<4) // in kilobytes
/*--------------------------------------------------------------------
 * 
 *  Module flags.
 *  These variables can be overridden at load time by the insmod
 *  command.
 *
 *--------------------------------------------------------------------*/

int         vid = PCI_VENDOR_ID_AMCC;
int         did = PCI_DEVICE_ID_AMCC_S5933;
/*
int vid = 0x10dc;
int did = 0x11;
*/

/*--------------------------------------------------------------------
 * 
 *  Static variables.
 *  These can only be seen by functions in this file.
 *
 *--------------------------------------------------------------------*/

/*
*   I think I have to rewrite the driver so that 
*   buffer_size, num_buffers, and total_buffer_size
*   are stored for each 'driver'
*/

static struct autobahn ab_dev[AB_MAX_DEVICES];  /* General data */
static u_short id_cnt = 0;                  /* ID counter */
static int ab_dev_cnt = 0;                      /* device counter */

static int buffer_size = BUFFER_SIZE;
static int num_buffers = 1;
static int total_buffer_size=0;
MODULE_PARM(buffer_size, "i");
MODULE_PARM(num_buffers, "i");
MODULE_PARM(total_buffer_size, "i");
MODULE_LICENSE("GPL");
MODULE_AUTHOR("NIST");
/* wait queues */

//static struct wait_queue *p2awq;
//static struct wait_queue *a2pwq;

/*--------------------------------------------------------------------
 * 
 *  Function prototypes
 *
 *--------------------------------------------------------------------*/

int         init_module (void);
void        cleanup_module (void);

static unsigned int ab_poll (struct file *, poll_table *);
static int  ab_ioctl (struct inode *, struct file *, unsigned int, unsigned long);
static int  ab_open (struct inode *, struct file *);
static int  ab_release (struct inode *, struct file *);
static int  ab_fasync (int , struct file *, int);
static int ab_mmap(struct file * file, struct vm_area_struct * vma);

static int  ab_init_dev (struct autobahn *);
static void ab_free_dev (struct autobahn *);
static void ab_irq_handler (int, void *, struct pt_regs *);
static void set_register(unsigned long ioaddr, unsigned long target);
static void get_register(unsigned long ioaddr, unsigned long *target);
static void printIOCTL(void);

/*--------------------------------------------------------------------
 *
 *  Linux kernel main entry point.
 *  For a definition of its fields see <linux/fs.h>.
 *
 *--------------------------------------------------------------------*/

//static struct file_operations ab_fops =
//{
//  NULL,      /* lseek()   */
//  NULL,      /* read()    */
//  NULL,      /* write()   */
//  NULL,      /* readdir() */
//  ab_poll,      /* poll()    */
//  ab_ioctl,      /* ioctl()   */
//  NULL,      /* mmap()    */
//  ab_open,      /* open()    */
//  NULL,                       /* flush()   */
//  ab_release,      /* release() */
//  NULL,      /* fsync()   */
//  ab_fasync,      /* fasync()  */
//  NULL,      /* check_media_change() */
//  NULL,      /* revalidate() */
//  NULL                        /* lock()    */
//};

static struct file_operations ab_fops = {
  open : ab_open,
  poll : ab_poll,
  ioctl: ab_ioctl,
  release : ab_release,
  mmap   : ab_mmap,
  fasync : ab_fasync
};

/*--------------------------------------------------------------------
 * 
 *  Standard module initialization function.
 *  This function scans the PCI bus looking for the right board.
 *
 *--------------------------------------------------------------------*/

int init_module (void) {
  extern int  vid, did, ab_dev_cnt;
  extern struct autobahn ab_dev[AB_MAX_DEVICES];

  struct pci_dev *pdev = NULL;
  //struct pci_dev *ab_pdev[AB_MAX_DEVICES];
  int         result = 0;
  int         result_init = 0;
  int         ab_if;
  unsigned long base0;

  pr_debug (AB_NAME ": >>> START <<<\n");
  ab_dev_cnt=0;
  ab_if=0;
  pr_debug(AB_NAME ": Looking for: vid: 0x%x\tdid: 0x%x\n",vid,did);
  /* loop over devices */
  if (pci_present()) {
    while ((pdev = pci_find_device((u_short) vid, (u_short) did, pdev))) {
      ab_dev[ab_if].pdev = pdev;
      pr_debug (AB_NAME 
                  ": init_mod: device found: IF %1.1d, BUS:%02x, FN:%02x\n",
      ab_if, pdev->bus->number, PCI_FUNC(pdev->devfn));
  
      /* major number allocation */
      if (ab_if == 0) {
        result = register_chrdev (AB_MAJOR, AB_NAME, &ab_fops);
        if (result) {
          printk (KERN_WARNING AB_NAME ": unable to get major number\n");
          return -EIO;
        }
      }
  
      /* initialize device */
      result_init = ab_init_dev (&ab_dev[ab_if]);
      base0 = pci_resource_start(pdev, 0);
//      pci_read_config_dword(pdev, PCI_BASE_ADDRESS_0, base0);
      pr_debug("base0: %lx\n",base0);
      ab_dev_cnt++;
      ab_if++;
    }
    if (ab_if == 0) {
      printk (KERN_WARNING AB_NAME ": pci error: %d: Can't find a card\n",result);
      return -ENODEV;  /* device not found */
    }
  } else {
    printk (KERN_WARNING AB_NAME ": pci error: %d: No PCI bus\n",result);
    return -ENODEV;
  }
  /* end of loop over devices */
  //printIOCTL();
  return result_init;
}

static void printIOCTL() {
  printk("ABIORESET:%x\n",ABIORESET);
  printk("ABIORESETADDON:%x\n",ABIORESETADDON);
  printk("ABIOSETLATENCY:%x\n",ABIOSETLATENCY);
  printk("ABIOSTART:%x\n",ABIOSTART);
  printk("ABIOSTOP:%x\n",ABIOSTOP);
  printk("ABIOCONFIGURE:%x\n",ABIOCONFIGURE);
  printk("ABIOGMCSR:%x\n",ABIOGMCSR);
  printk("ABIOGINTCSR:%x\n",ABIOGINTCSR);
  printk("ABIOGINFOADDR:%x\n",ABIOGINFOADDR);
}
/*--------------------------------------------------------------------
 * 
 *  Standard module release function.
 *
 *--------------------------------------------------------------------*/

void cleanup_module (void) {
    extern struct autobahn ab_dev[AB_MAX_DEVICES];
    extern int ab_dev_cnt;
    int ab_if;

    for (ab_if=0;  ab_if < ab_dev_cnt; ab_if++) {
      ab_free_dev (&ab_dev[ab_if]);
    }
    pr_debug (AB_NAME ": >>> STOP <<<\n");
}

/*--------------------------------------------------------------------
 * 
 *  Standard open() entry point.
 *  Very simple for the moment.
 *
 *--------------------------------------------------------------------*/

static int
ab_open (struct inode *inode, struct file *file)
{
    struct ab_procdata *pd;
    pr_debug (AB_NAME ": open\n");

    /* allocate private data */

    pd = kmalloc (sizeof (struct ab_procdata), GFP_KERNEL);
    if (pd == NULL) {
      printk (KERN_WARNING AB_NAME ": open: no memory for private data\n");
      return -ENOMEM;
    }

    memset (pd, 0, sizeof (struct ab_procdata));

    /* generate unique ID */
    pd->pd_id = (id_cnt << 16) | (current->pid);
    pd->pd_if = MINOR(inode->i_rdev);
    pr_debug (AB_NAME ": open: ID: %d I/F: %d\n", pd->pd_id, pd->pd_if);

    file->private_data = pd;  /* save information */

    MOD_INC_USE_COUNT;
    return 0;
}

/*--------------------------------------------------------------------
 * 
 *  Standard release() entry point.
 *  This function is called by the close() system call.
 *
 *--------------------------------------------------------------------*/

static int
ab_release (struct inode *inode, struct file *file)
{
    struct ab_procdata *pd = file->private_data;

    pr_debug (AB_NAME ": close\n");
	ab_fasync(-1, file, 0);

    if (pd) {
      pr_debug (AB_NAME ": close: ID: %d\n", pd->pd_id);

  /* bigphys buffer */
//  I will free memory here when I add user configurable memory allocation
/*
  if (pd->pd_virt)
      bigphysarea_free (pd->pd_virt, pd->pd_size);
*/
      kfree (pd);
      file->private_data = NULL;
    }

    MOD_DEC_USE_COUNT;
    return 0;
}

/*--------------------------------------------------------------------
 * 
 *  Standard poll() entry point.
 *
 *--------------------------------------------------------------------*/

static unsigned int ab_poll (struct file *file, poll_table *poll_table) {
  unsigned int mask = 0;
  extern struct autobahn ab_dev[AB_MAX_DEVICES];
  struct ab_procdata *pd = file->private_data;
  int ab_if = pd->pd_if;
  struct autobahn  dev = ab_dev[ab_if];
  struct ab_driver *drv = dev.ab_drv;

  if (drv->irq_flag) {
    drv->irq_flag = 0;
    mask |= POLLIN | POLLRDNORM;  /* readable */ 
  }
  return mask;
}

/*--------------------------------------------------------------------
 * 
 *  fasync() entry point.
 *
 *--------------------------------------------------------------------*/

static int ab_fasync (int fd, struct file *file, int mode){
  extern struct autobahn ab_dev[AB_MAX_DEVICES];
  struct ab_procdata *pd = file->private_data;
  
  return fasync_helper(fd, file, mode, &(ab_dev[pd->pd_if].async_queue));
}

/*--------------------------------------------------------------------
 * 
 *  mmap() entry point.
 *
 *--------------------------------------------------------------------*/

static int ab_mmap(struct file * file, struct vm_area_struct * vma) {
  struct ab_procdata *pd = file->private_data;
  int ab_if = pd->pd_if;
  struct autobahn  dev = ab_dev[ab_if];
  struct ab_driver *drv = dev.ab_drv;
  int err;
  u_long size = vma->vm_end - vma->vm_start;
  u_long offset;

  offset = (u_long)drv->pd_phys;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,25)
  vma->vm_offset = offset;
#else
  vma->vm_pgoff = offset >> PAGE_SHIFT;
#endif

  if( (offset|size) & ~PAGE_MASK )
    return -EFAULT;
//  pr_debug((AB_NAME "ab: params for remap: vm_start 0x%0x offset 0x%0x\n",
//		(u_int)vma->vm_start, (u_int)offset));
  if ( (err=remap_page_range(vma->vm_start, offset, size, vma->vm_page_prot)) )
    return err;

  vma->vm_file = file;
  //get_file(file);
      
  return 0;
}

/*--------------------------------------------------------------------
 * 
 *  Standard ioctl() entry point.
 *
 *--------------------------------------------------------------------*/

static int
ab_ioctl (struct inode *inode, struct file *file, unsigned int cmd,
    unsigned long arg)
{
    extern struct autobahn ab_dev[AB_MAX_DEVICES];

    int ab_if = MINOR(inode->i_rdev);
    struct autobahn  dev = ab_dev[ab_if];
    struct ab_driver *drv = dev.ab_drv;
    int         ret = 0;
    _mcsr mcsr;
    _icsr intcsr;
    unsigned long *tempaddr;
    unsigned long long *cpu_time;
    unsigned long ioaddr;
    ioaddr = (unsigned long)(dev.ab_opregs_p);
    /*
    pr_debug (AB_NAME ": ioctl: %08lx dev: %08lx, drv: %08lx, ab_if: %x\n", 
               (unsigned long) cmd, (unsigned long) &dev, 
               (unsigned long) drv, ab_if);
	       */
    //pr_debug (AB_NAME ": ioctl: cmd: %08lx, arg: %08lx\n", (unsigned long)cmd, arg);
    switch (cmd) {

  /* general */

    case ABIORESET:    /* reset the whole board */
      pr_debug (AB_NAME ": ABIORESET\n");

      mcsr.reg = 0;
      mcsr.p2a_reset = 1;
      mcsr.a2p_reset = 1;
      set_register(ioaddr + AMCC_OP_REG_MCSR, mcsr.reg);

      intcsr.reg = 0x3C000000; // set FIFO advance control no endian conversion 
       
      set_register(ioaddr + AMCC_OP_REG_INTCSR, intcsr.reg);
      set_register(ioaddr + AMCC_OP_REG_MWTC, 0);
      set_register(ioaddr + AMCC_OP_REG_MRTC, 0);
      drv->counter = 0;
      drv->irq_flag = 0;
      drv->buffer_number = 0;

      break;


    case ABIORESETADDON:  /* reset the add-on */
      // Should I read mcsr instead?
      mcsr.reg = 0;
      mcsr.ao_reset = 1;
      set_register(ioaddr + AMCC_OP_REG_MCSR, mcsr.reg);
      udelay (20UL);
      mcsr.ao_reset = 0;
      set_register(ioaddr + AMCC_OP_REG_MCSR, mcsr.reg);
      break;

    case ABIOSETLATENCY:  /* set PCI latency timer */
      if (arg < 0x20 || arg > 0xF8) {
        printk (KERN_WARNING AB_NAME ": ioctl: invalid PCI latency %02x\n", (unsigned int) arg);
        return -EINVAL;
      }
      /* Look at me */
      pci_write_config_byte (dev.pdev, PCI_LATENCY_TIMER, arg);
    break;
    
    case ABIOSETNUMBUFFERS:  /* set the number of write buffers */
      num_buffers = arg;
      if (num_buffers * buffer_size > total_buffer_size) {
	return -EINVAL;
      }
    break;

    case ABIOSETBUFFERSIZE:  /* set the buffer_size */
      buffer_size = arg;
      if (num_buffers * buffer_size > total_buffer_size) {
	return -EINVAL;
      }
    break;
       
    case ABIOSTART: /* START THE DAQ */
      {
        unsigned int *temp;
		unsigned long mrtc_result;
		unsigned int mwtc_size = buffer_size<<10;
		unsigned long mwar_start = (unsigned long)(drv->pd_phys)
			                        + mwtc_size * drv->buffer_number;
        //struct ab_procdata *pd = file->private_data;
        printk("ABIOSTART: mwar_start: %08lx\t ",mwar_start);

        // Reset FIFO's 
        mcsr.reg = 0;
        mcsr.p2a_reset = 1;
        mcsr.a2p_reset = 1;
        set_register(ioaddr + AMCC_OP_REG_MCSR, mcsr.reg);
        mcsr.reg = 0;
        // Setup read FIFO 
        set_register(ioaddr + AMCC_OP_REG_MRAR, (unsigned long)(drv->pd_phys_read));
        set_register(ioaddr + AMCC_OP_REG_MRTC, (1<<2));
        temp = (unsigned int *) (drv->pd_virt_read);
        *temp = arg;
        printk("p2a content: %x\n",  *temp);
        // Setup write FIFO 
        set_register(ioaddr + AMCC_OP_REG_MWAR, mwar_start);
        set_register(ioaddr + AMCC_OP_REG_MWTC, mwtc_size);
        // Enable the WTC interrupt
        intcsr.reg = 0x3C000000; 
        intcsr.wr_compl_en = 1; 
        set_register(ioaddr + AMCC_OP_REG_INTCSR, intcsr.reg);
        // Start add-on to PCI
        mcsr.wtc_enable = 1;
        set_register(ioaddr + AMCC_OP_REG_MCSR, mcsr.reg);
        mcsr.rtc_enable = 1;
        mcsr.wtc_w_prio = 1;
        mcsr.rtc_r_prio = 1;
        mcsr._s._wtc_fifo_mng = 1;
        set_register(ioaddr + AMCC_OP_REG_MCSR, mcsr.reg);

        get_register(ioaddr + AMCC_OP_REG_MRTC, &mrtc_result);
        printk("mrtc_result: %lx\n", mrtc_result);
        while (mrtc_result>0)
          get_register(ioaddr + AMCC_OP_REG_MRTC, &mrtc_result);
        // Disable transfer
        printk("mrtc_result: %lx\n", mrtc_result);
        mcsr.rtc_enable = 0;
        set_register(ioaddr + AMCC_OP_REG_MCSR, mcsr.reg);
      }
      break;        
    case ABIOSTOP:    /* stop the DAQ  */
      mcsr.reg = 0;
      set_register(ioaddr + AMCC_OP_REG_MCSR, mcsr.reg);

      break;
    case ABIOCONFIGURE: {  //  Send the configuration word
        unsigned int *temp;
        unsigned long ltemp; 
        unsigned long mrtc_result;
        printk("ABIOCONFIGURE START: %lx\n",arg);
        // Reset FIFO's 
        get_register(ioaddr + AMCC_OP_REG_MCSR, &ltemp); 
        mcsr.reg = ltemp;
        mcsr.p2a_reset = 1;
        printk("ABIOCONFIGURE reset p2a\n");
        set_register(ioaddr + AMCC_OP_REG_MCSR, mcsr.reg);
        printk("ABIOCONFIGURE: set MRAR\n" );
        set_register(ioaddr + AMCC_OP_REG_MRAR, (unsigned long)(drv->pd_phys_read));

        printk("ABIOCONFIGURE: setup message\n");
        temp = (unsigned int *) (drv->pd_virt_read);
        *temp = arg;
        printk("p2a content: %x\n", *temp);
        printk("ABIOCONFIGURE: set MRTC\n");
        set_register(ioaddr + AMCC_OP_REG_MRTC, (1<<2));

        get_register(ioaddr + AMCC_OP_REG_MRTC, &mrtc_result);
        printk("mrtc_result: %lx\n", mrtc_result);
        // Enable transfer of data
        //mcsr.reg = 0;
        mcsr.rtc_enable = 1;
        set_register(ioaddr + AMCC_OP_REG_MCSR, mcsr.reg);
        // Wait until transfer is done
        get_register(ioaddr + AMCC_OP_REG_MRTC, &mrtc_result);
        printk("mrtc_result: %lx\n", mrtc_result);
        while (mrtc_result>0)
          get_register(ioaddr + AMCC_OP_REG_MRTC, &mrtc_result);
        // Disable transfer
        printk("mrtc_result: %lx\n", mrtc_result);
        get_register(ioaddr + AMCC_OP_REG_MCSR, &ltemp);
        mcsr.reg = ltemp;
        mcsr.rtc_enable = 0;
        set_register(ioaddr + AMCC_OP_REG_MCSR, mcsr.reg);
        printk("ABIOCONFIGURE STOP: %lx\n",arg);
      }
      break;
    case ABIOGMCSR:  
      tempaddr = (unsigned long *)arg; 
      get_register(ioaddr + AMCC_OP_REG_MCSR, tempaddr); 
	  printk("mcsr: %lx\n",*tempaddr);
      break;
    case ABIOGINTCSR:
      tempaddr = (unsigned long *)arg; 
      get_register(ioaddr + AMCC_OP_REG_INTCSR, tempaddr); 
      break;
    case ABIOGINFOADDR:
      tempaddr = (unsigned long *)arg; 
      *tempaddr = (unsigned long)(dev.ab_drv_p);
      break;
    case ABIOGA2PBUFFER:
      tempaddr = (unsigned long *)arg; 
      *tempaddr = (unsigned long) drv->pd_phys;
      break;
    case ABIOGP2ABUFFER:
      tempaddr = (unsigned long *)arg; 
      *tempaddr = (unsigned long) drv->pd_phys_read;
      break;
    case ABIOGMWTC:
      tempaddr = (unsigned long *)arg; 
      get_register(ioaddr + AMCC_OP_REG_MWTC, tempaddr); 
      break;
    case ABIOGMRTC:
      tempaddr = (unsigned long *)arg; 
      get_register(ioaddr + AMCC_OP_REG_MRTC, tempaddr); 
      break;
    case ABIOGBUFFNUMBER:
      tempaddr = (unsigned long *)arg; 
      *tempaddr = drv->buffer_number; 
      break;
    case ABIOGNUMBUFFERS:
      tempaddr = (unsigned long *)arg; 
      *tempaddr = num_buffers; 
      break;
    case ABIOGCOUNTER:
      tempaddr = (unsigned long *)arg; 
      *tempaddr = drv->counter; 
      break;
    case ABIOGBUFFERSIZE:
      tempaddr = (unsigned long *)arg; 
      *tempaddr = drv->buffer_size;
      break;
    case ABIOGCPUTIME:
      cpu_time = (unsigned long long*)arg; 
      *cpu_time = drv->drv_time;
      break;
    /* End of ioctls */
    default:
      ret = -EINVAL;
      break;
    }

    return ret;
}


/*--------------------------------------------------------------------
 * 
 *  Device initialization.
 *  This function initializes the board. It requests an interrupt and
 *  maps the operation registers in kernel space. The major number
 *  has already been allocated by the module init function.
 *
 *--------------------------------------------------------------------*/

static int
ab_init_dev (struct autobahn *dev)
{
    u32         data;
    //u_long      ret;
    //u_int       mask;
    _mcsr mcsr;
    _icsr intcsr;
    u_long ioaddr;
    unsigned long size_kb, size_page;
    void *buf;
    /* Need to map upper memory Allocate shared data page */

    size_kb = sizeof(struct ab_driver) >> 10;
    if ( sizeof(struct ab_driver) > (size_kb<<10)) size_kb++;
    size_page = sizeof(struct ab_driver) >> PAGE_SHIFT;
    if ( sizeof(struct ab_driver) > (size_page<<PAGE_SHIFT)) size_page++;
    //dev->ab_drv_p = (caddr_t) allocator_allocate_dma(size_kb, GFP_KERNEL);
    dev->ab_drv = (void *) bigphysarea_alloc_pages(size_page, 1, GFP_KERNEL);
    if (dev->ab_drv == NULL) {
      printk (KERN_WARNING AB_NAME 
        ": init_mod: no memory for ab_driver structure\n");
      return -ENOMEM;
    }
    /*
    dev->ab_drv_p = (caddr_t) bigphysarea_alloc_pages(size_page, 1, GFP_KERNEL);
    if (dev->ab_drv_p == NULL) {
      printk (KERN_WARNING AB_NAME 
        ": init_mod: no memory for ab_driver structure\n");
      return -ENOMEM;
    }
    */ 
    pr_debug (AB_NAME ": init_mod: ab_drv_p: %08lx\n", (unsigned long)(dev->ab_drv_p));
    pr_debug (AB_NAME ": init_mod: high_memory: %08lx\n", (unsigned long)high_memory);
    pr_debug (AB_NAME ": init_mod: IRQ num: %d\n", (int)(dev->pdev->irq));
      
    /* Caution: We are not sure if this still works in 2.2 */
    /*
    dev->ab_drv = (struct ab_driver *) 
      ioremap ((u_long) dev->ab_drv_p,sizeof (struct ab_driver));
      */
    dev->ab_drv_p = (caddr_t) virt_to_phys(dev->ab_drv);
    memset (dev->ab_drv, 0, sizeof (struct ab_driver));

    pr_debug (AB_NAME ": init_mod: dev->ab_drv: %08lx\n", 
              (unsigned long)(dev->ab_drv));
    pr_debug (AB_NAME ": init_mod: ab_drv_p: %08lx\n", (unsigned long)(dev->ab_drv_p));
    pr_debug (AB_NAME ": init_mod: high_memory: %08lx\n", (unsigned long)high_memory);
    pr_debug (AB_NAME ": init_mod: IRQ num: %d\n", (int)(dev->pdev->irq));
      
      
    /* 
     * IRQ facilities setup.
     * Register the irq with the system.
     */
/*
    if (request_irq (dev->pdev->irq, ab_irq_handler, 
         SA_SHIRQ, AB_NAME, NULL)) {
*/
    if (request_irq (dev->pdev->irq, ab_irq_handler, 
         SA_SHIRQ, AB_NAME, dev)) {
      printk (KERN_WARNING AB_NAME ": can't install handler for irq %d\n",
        dev->pdev->irq);
      ab_free_dev (dev);
      return -EIO; 
    }

    /* Enable bus master */
    //pci_set_master(dev->pdev);

    /* 
     * Retrieve from PCI_BASE_ADDRESS_0 the address of the S5933 
     * operations registers and check for memory mapped operation 
     * registers.
     */

    //data = (dev->pdev->base_address[0] & PCI_BASE_ADDRESS_MEM_MASK);
    pci_read_config_dword(dev->pdev, PCI_BASE_ADDRESS_0, &data);
    data &= PCI_BASE_ADDRESS_IO_MASK;
    dev->ab_opregs_p = (caddr_t) data;
    //dev->ab_opregs = (struct s5933_opreg *) ioremap (data, AMCC_OP_REG_SIZE);

    /* board reset */

    mcsr.reg = 0;
    ioaddr = (unsigned long) (dev->ab_opregs_p);
    mcsr.p2a_reset = 1;
    mcsr.a2p_reset = 1;
    set_register(ioaddr + AMCC_OP_REG_MCSR, mcsr.reg);
    // set FIFO advance, no endian conv., disable interrupts
    intcsr.reg = 0x3C000000;
     
    set_register(ioaddr + AMCC_OP_REG_INTCSR, intcsr.reg);
    set_register(ioaddr + AMCC_OP_REG_MWTC, 0);
    set_register(ioaddr + AMCC_OP_REG_MRTC, 0);

    /* some debug info */

    pr_debug (AB_NAME ": init: MCSR: %08lx\n", 
              (unsigned long)(mcsr.reg));
    pr_debug (AB_NAME ": init: ICSR: %08lx\n", 
              (unsigned long)(intcsr.reg));
    // Allocate memory for the A2P and P2A transfers
    //  Eventually I should move this to an IOCTL command
    if (buffer_size * num_buffers > total_buffer_size) {  
      total_buffer_size = buffer_size * num_buffers;
    } 

    //printk("pd_phys before: %p\n",( dev->ab_drv->pd_phys) );
 
    //buf = (unsigned long) allocator_allocate_dma((unsigned long)(total_buffer_size), GFP_KERNEL); 
    dev->ab_drv->pd_virt = bigphysarea_alloc_pages((unsigned long)(total_buffer_size)>>(PAGE_SHIFT-10), 1, GFP_KERNEL); 
    if (dev->ab_drv->pd_virt == NULL) {
      printk (KERN_WARNING AB_NAME ": open: no memory for write buffer\n");
      return -ENOMEM;
    }
    //dev->ab_drv->pd_phys = (caddr_t) buf;
    //dev->ab_drv->pd_phys = (void *) buf;
    //dev->ab_drv->pd_virt = ioremap(dev->ab_drv->pd_phys, (total_buffer_size)<<10);
    dev->ab_drv->pd_phys = (caddr_t) virt_to_phys(dev->ab_drv->pd_virt);

    pr_debug("pd_phys: %p  virt: %p\n",( dev->ab_drv->pd_phys),
            dev->ab_drv->pd_virt );
 
    //buf = allocator_allocate_dma (P2A_SIZE, GFP_KERNEL); 
    buf = bigphysarea_alloc_pages (P2A_SIZE>>(PAGE_SHIFT-10), 1, GFP_KERNEL); 
    if (buf == 0) {
      printk (KERN_WARNING AB_NAME ": open: no memory for read buffer\n");
      return -ENOMEM;
    }
    dev->ab_drv->pd_virt_read = (caddr_t) buf;
    //dev->ab_drv->pd_virt_read = ioremap(buf,(P2A_SIZE)<<10);
    dev->ab_drv->pd_phys_read = (caddr_t) virt_to_phys(buf);
    pr_debug (AB_NAME ": init: pd_phys_read: %p  virt: %p\n", 
              (dev->ab_drv->pd_phys_read), 
	      (dev->ab_drv->pd_virt_read) );
    dev->ab_drv->counter = 0;
    dev->ab_drv->irq_flag = 0;
	dev->ab_drv->buffer_number = 0;
    dev->ab_drv->buffer_size = buffer_size;
	pr_debug("num_buffers: %d\tbuffer_size: %d Kb\n",num_buffers, buffer_size);
    return 0;
}

/*--------------------------------------------------------------------
 * 
 *  Device finalization.
 *  It selectively frees all the resources allocated by the driver.
 *
 *--------------------------------------------------------------------*/

static void
ab_free_dev (struct autobahn *dev)
{
    _mcsr mcsr;
    //_icsr intcsr;
    unsigned long ioaddr;
    /* leave the board in a clean state */
    // should I get mcsr?
    mcsr.reg = 0;
    ioaddr = (unsigned long)(dev->ab_opregs_p);
    mcsr.p2a_reset = 1;
    mcsr.a2p_reset = 1;
    set_register(ioaddr + AMCC_OP_REG_MCSR, mcsr.reg);
    set_register(ioaddr + AMCC_OP_REG_MCSR, 0);
    set_register(ioaddr + AMCC_OP_REG_INTCSR, 0);
    set_register(ioaddr + AMCC_OP_REG_MWTC, 0);
    set_register(ioaddr + AMCC_OP_REG_MRTC, 0);

    /* driver status  */

    if (dev->ab_drv_p) {
      //iounmap ((dev->ab_drv)->pd_virt);
      //allocator_free_dma((unsigned long)((dev->ab_drv)->pd_phys));
      bigphysarea_free_pages(((dev->ab_drv)->pd_virt));
      //iounmap ((dev->ab_drv)->pd_virt_read);
      //allocator_free_dma((unsigned long)((dev->ab_drv)->pd_phys_read));
      bigphysarea_free_pages (((dev->ab_drv)->pd_virt_read));
      //iounmap (dev->ab_drv);
      //allocator_free_dma((unsigned long)(dev->ab_drv_p));
      bigphysarea_free_pages ((void *)dev->ab_drv);
    }

    /* IRQ & major number */
    free_irq (dev->pdev->irq, dev);
    unregister_chrdev (AB_MAJOR, AB_NAME);
}

/*--------------------------------------------------------------------
 * 
 *  Interrupt handler.
 *
 *--------------------------------------------------------------------*/

static void
ab_irq_handler (int irq, void *device, struct pt_regs *regs)
{
  extern struct autobahn ab_dev[AB_MAX_DEVICES];
  int ab_if = 0;
  u_llong d_time;
  unsigned long ioaddr, temp;
  _icsr intcsr;
  _mcsr mcsr;
  unsigned int mwtc_size = buffer_size<<10;
  unsigned long mwar_start = (unsigned long)(ab_dev[ab_if].ab_drv)->pd_phys;
  //struct ab_procdata *pd = file->private_data;

  /* first of all we get the clock */

  d_time = getclock ();

  /* then we check :-) */

  for ( ab_if=0; ab_if<ab_dev_cnt; ab_if++) {
    if (irq == ab_dev[ab_if].pdev->irq) {
      break;
    } else {
      if (ab_if+1==ab_dev_cnt) {
        printk (KERN_WARNING AB_NAME ": irq: spurious interrupt!\n");
        return;
      }
    }
  }
  //pr_debug (AB_NAME ": interrupt: irq: %d\n", irq);

  /* get driver status and store interrupt time */

  (ab_dev[ab_if].ab_drv)->drv_time = d_time;
  //printk(AB_NAME ": interrupt d_time: %lx\n",d_time);
  ioaddr = (unsigned long)(ab_dev[ab_if].ab_opregs_p);

  /* read interrupt code */
  get_register(ioaddr + AMCC_OP_REG_INTCSR, &temp); 
  intcsr.reg = temp;
  if (!(intcsr.int_asserted))
    return;
  //printk(AB_NAME ": interrupt d_time: %Lx\n",d_time);
  //printk(AB_NAME ": interrupt intcsr: %lx\n",temp);

  /* local copy */
  (ab_dev[ab_if].ab_drv)->drv_icsr = intcsr.reg;
  /*
  pr_debug (AB_NAME ": irq: ICSR: %08lx\n", 
      (unsigned long) ((ab_dev[ab_if].ab_drv)->drv_icsr));
  */
  get_register(ioaddr + AMCC_OP_REG_MCSR, &temp); 
  mcsr.reg = temp;
  //pr_debug (AB_NAME ": interrupt: MCSR: %08lx\n", (unsigned long)mcsr.reg);

  (ab_dev[ab_if].ab_drv)->counter++; 
  (ab_dev[ab_if].ab_drv)->irq_flag = 1; 
  (ab_dev[ab_if].ab_drv)->buffer_number++; 
  (ab_dev[ab_if].ab_drv)->buffer_number %= num_buffers; 
  
  /* 
   *  Disable bus-mastering if needed.
   *  We do it in a selective way.
   */

  if ((ab_dev[ab_if].ab_drv)->drv_icsr & ICSR32_DMA_ABORT) {

  /* PCI bus error */
    mcsr.wtc_enable = 0;
    mcsr.rtc_enable = 0;
    set_register(ioaddr + AMCC_OP_REG_MCSR, mcsr.reg); 
    
    //(ab_dev[ab_if].ab_opregs)->wtc_enable = 0;     /* stop transfer */
    //(ab_dev[ab_if].ab_opregs)->rtc_enable = 0;

    intcsr.wr_compl_en = 0;
    intcsr.rd_compl_en = 0;
    set_register(ioaddr + AMCC_OP_REG_INTCSR, (intcsr.reg)); 
    //(ab_dev[ab_if].ab_opregs)->wr_compl_en = 0;    /* disable interrupts */
    //(ab_dev[ab_if].ab_opregs)->rd_compl_en = 0;
  } else {

  /* normal end of transfer */

    if ((ab_dev[ab_if].ab_drv)->drv_icsr & ICSR32_RT_COMPL_INT) {
      mcsr.rtc_enable = 0;
      set_register(ioaddr + AMCC_OP_REG_MCSR, mcsr.reg); 
      intcsr.rd_compl_en = 0;
      set_register(ioaddr + AMCC_OP_REG_INTCSR, (intcsr.reg)); 
      //(ab_dev[ab_if].ab_opregs)->rtc_enable = 0;
      //(ab_dev[ab_if].ab_opregs)->rd_compl_en = 0;
    }
    if ((ab_dev[ab_if].ab_drv)->drv_icsr & ICSR32_WT_COMPL_INT) {
      mcsr.wtc_enable = 0;
      set_register(ioaddr + AMCC_OP_REG_MCSR, mcsr.reg); 
      intcsr.wr_compl_en = 0;
      set_register(ioaddr + AMCC_OP_REG_INTCSR, (intcsr.reg)); 
      //(ab_dev[ab_if].ab_opregs)->wtc_enable = 0;
      //(ab_dev[ab_if].ab_opregs)->wr_compl_en = 0;
    }
  }

  /* clear the interrupt source */

  intcsr.int_code = ~0;
  set_register(ioaddr + AMCC_OP_REG_INTCSR, (intcsr.reg)); 
  //(ab_dev[ab_if].ab_opregs)->int_code = ~0;

  get_register(ioaddr + AMCC_OP_REG_INTCSR, &temp); 
  intcsr.reg = temp;
  //pr_debug (AB_NAME ": interrupt: MCSR: %08lx\n", (unsigned long)temp);

  /* 
   *  Handle interrupt
   */

  if ((ab_dev[ab_if].ab_drv)->drv_icsr & ICSR32_WT_COMPL_INT) {

    /* write (Add-on -> PCI) complete */

    //set_bit (AB_FLAG_A2PDONE, &((ab_dev[ab_if].ab_drv)->drv_lock));
        // Setup write FIFO 
        //temp = (unsigned long)(ab_dev[ab_if].ab_drv)->pd_phys;
        mwar_start += mwtc_size * (ab_dev[ab_if].ab_drv)->buffer_number;
        set_register(ioaddr + AMCC_OP_REG_MWAR, mwar_start);
        set_register(ioaddr + AMCC_OP_REG_MWTC, mwtc_size);
        // Enable the WTC interrupt
        intcsr.wr_compl_en = 1; 
        set_register(ioaddr + AMCC_OP_REG_INTCSR, intcsr.reg);

        // Start add-on to PCI
        mcsr.wtc_enable = 1;
        set_register(ioaddr + AMCC_OP_REG_MCSR, mcsr.reg);
        //mcsr.rtc_enable = 1;
        mcsr.rtc_enable = 0;
        mcsr.wtc_w_prio = 1;
        mcsr.rtc_r_prio = 1;
        mcsr._s._wtc_fifo_mng = 1;
        set_register(ioaddr + AMCC_OP_REG_MCSR, mcsr.reg);
       
    //wake_up_interruptible (&a2pwq);
  }

  if ((ab_dev[ab_if].ab_drv)->drv_icsr & ICSR32_RT_COMPL_INT) {

    /* read (PCI -> Add-on) complete */

    //set_bit (AB_FLAG_P2ADONE, &((ab_dev[ab_if].ab_drv)->drv_lock));
    //wake_up_interruptible (&p2awq);
  }
  if (ab_dev[ab_if].async_queue) {
    kill_fasync(&(ab_dev[ab_if].async_queue), SIGIO, POLL_IN);
  }
}

static void set_register(unsigned long ioaddr, unsigned long target) {
  //printk("Setting: %lx: %lx\n",ioaddr, target);
  outl(target, ioaddr);  // stupid parameter list order  content,addr
}                                                                               

static void get_register(unsigned long ioaddr, unsigned long *target) {
  *target = inl(ioaddr);
}                                                                               
/* end of code */
