#ifndef _amcc_H
#define _amcc_H

#ifndef _LINUX_TYPES_H
# include <linux/types.h>
#endif

/*-- Defines ---------------------------------------------------------*/

#define PCI_DEVICE_ID_AMCC_S5933        0x4750
#define AMCC_OP_REG_SIZE		(sizeof (struct s5933_opreg))

/* these definitions are for faster handling of critical situations */

#define MCSR_RTC_ENABLE		0x4	/* 3 bit */
#define MCSR_RTC_FIFO_MNG	0x2
#define MCSR_RTC_RD_PRIO	0x1

#define MCSR_WTC_ENABLE		0x4	/* 3 bit */
#define MCSR_WTC_FIFO_MNG	0x2
#define MCSR_WTC_WR_PRIO	0x1

#define ICSR_TARGET_ABORT       0x20	/* 6 bit */
#define ICSR_MASTER_ABORT       0x10
#define ICSR_RT_COMPL_INT       0x08
#define ICSR_WT_COMPL_INT       0x04

/* 32-bit handling of MCSR */

#define MCSR32_GLOBAL_RESET		0x0F000000L	/* reset */
#define  MCSR32_MB_RESET		0x08000000L
#define  MCSR32_A2P_FIFO_RESET		0x04000000L
#define  MCSR32_P2A_FIFO_RESET		0x02000000L
#define  MCSR32_ADD_ON_RESET		0x01000000L

#define MCSR32_READ_TR_ENABLE		0x00004000L	/* read transfer control */
#define MCSR32_READ_IF_4_EMPTY		0x00002000L
#define MCSR32_READ_PRIORITY		0x00001000L

#define MCSR32_WRITE_TR_ENABLE		0x00000400L	/* write transfer control */
#define MCSR32_WRITE_IF_4_FILLED	0x00000200L
#define MCSR32_WRITE_PRIORITY		0x00000100L

#define MCSR32_A2P_TC_ZERO		0x00000080L	/* status */
#define MCSR32_P2A_TC_ZERO		0x00000040L

#define MCSR32_A2P_FIFO_EMPTY		0x00000020L
#define MCSR32_A2P_FIFO_HALF		0x00000010L
#define MCSR32_A2P_FIFO_FULL		0x00000008L
#define MCSR32_P2A_FIFO_EMPTY		0x00000004L
#define MCSR32_P2A_FIFO_HALF		0x00000002L
#define MCSR32_P2A_FIFO_FULL		0x00000001L

/* 32-bit handling of ICSR */

#define ICSR32_INT_ASSERTED		0x00800000L
#define ICSR32_TARGET_ABORT		0x00200000L
#define ICSR32_MASTER_ABORT		0x00100000L
#define  ICSR32_DMA_ABORT		(ICSR32_TARGET_ABORT | ICSR32_MASTER_ABORT)

#define ICSR32_RT_COMPL_INT		0x00080000L
#define ICSR32_WT_COMPL_INT		0x00040000L

#define ICSR32_IN_MB_INT		0x00020000L
#define ICSR32_OUT_MB_INT		0x00010000L

/****************************************************************************/
/*   AMCC Operation Register Offsets                                        */
/****************************************************************************/

#define AMCC_OP_REG_OMB1         0x00
#define AMCC_OP_REG_OMB2         0x04
#define AMCC_OP_REG_OMB3         0x08 
#define AMCC_OP_REG_OMB4         0x0c
#define AMCC_OP_REG_IMB1         0x11
#define AMCC_OP_REG_IMB2         0x14
#define AMCC_OP_REG_IMB3         0x18 
#define AMCC_OP_REG_IMB4         0x1c
#define AMCC_OP_REG_FIFO         0x20
#define AMCC_OP_REG_MWAR         0x24
#define AMCC_OP_REG_MWTC         0x28
#define AMCC_OP_REG_MRAR         0x2c
#define AMCC_OP_REG_MRTC         0x30
#define AMCC_OP_REG_MBEF         0x34
#define AMCC_OP_REG_INTCSR       0x38
#define AMCC_OP_REG_MCSR         0x3c
#define AMCC_OP_REG_MCSR_NVDATA  (AMCC_OP_REG_MCSR + 2) /* Data in byte 2 */
#define AMCC_OP_REG_MCSR_NVCMD   (AMCC_OP_REG_MCSR + 3) /* Command in byte 3 */

    typedef union {		/* Interrupt Control/Status Register */
	struct {
	    u_int       _intsel:16;	/* Interrupt selection */
	    u_int       _intcode:6;	/* Interrupt code */
	    u_int       :1;
	    u_int       _interrupt:1;	/* Interrupt asserted */
	    u_int       _fectl:8;	/* FIFO and Endian control */
	} _t;
	struct {
    		/* selection */
	    u_int       _isel_omb_byte:2;	/* OMB byte # */
	    u_int       _isel_omb_num:2;	/* OMB # */
	    u_int       _isel_omb_en:1;		/* Enable int. */
	    u_int       :3;
	    u_int       _isel_imb_byte:2;	/* IMB byte # */
	    u_int       _isel_imb_num:2;	/* IMB # */
	    u_int       _isel_imb_en:1;		/* Enable int. */
	    u_int       :1;
	    u_int       _isel_wtc:1;	/* Int. on write complete */
	    u_int       _isel_rtc:1;	/* Int. on read complete */
		/* code */
	    u_int       _int_omb:1;	/* OMB interrupt */
	    u_int       _int_imb:1;	/* IMB interrupt */
	    u_int       _int_wtc:1;	/* write compl. interrupt */
	    u_int       _int_rtc:1;	/* read compl. interrupt */
	    u_int       _int_mabort:1;	/* Master abort interrupt */
	    u_int       _int_tabort:1;	/* Target abort interrupt */
	    u_int       :1;
	    u_int       _int_on:1;	/* Interrupt asserted */
		/* FIFO & endian control */
	    u_int       _endian_conv:2;
	    u_int       _fctl_pci_adv:2;
	    u_int       _fctl_ao_adv:2;
	    u_int       _fctl_a2p_toggle:1;
	    u_int       _fctl_p2a_toggle:1;
	} _s;
	__u32       reg;
    } _icsr;

#define wr_compl_en	_s._isel_wtc
#define rd_compl_en	_s._isel_rtc
#define omb_en		_s._isel_omb_en
#define omb_num		_s._isel_omb_num
#define omb_byte	_s._isel_omb_byte
#define imb_en		_s._isel_imb_en
#define imb_num		_s._isel_imb_num
#define imb_byte	_s._isel_imb_byte
#define int_wr_compl	_s._int_wtc
#define int_rd_compl	_s._int_rtc
#define int_m_abort	_s._int_mabort
#define int_t_abort	_s._int_tabort
#define int_asserted	_t._interrupt
#define int_code	_t._intcode
#define int_selection	_t._intsel


    typedef union {		/* Bus Master Control Status register */
	struct {
	    u_int       _status:8;	/* FIFO status */
	    u_int       _wtc:3;		/* Write Transfer Control */
	    u_int	:1;
	    u_int	_rtc:3;		/* Read Transfer Control */
	    u_int	:1;
	    u_int	_nv_bus:8;	/* NV address/data */
	    u_int	_reset:4;	/* Reset Controls */
	    u_int	:1;
	    u_int	_nv_ctl:3;	/* NV Control */
	} _t;
	struct {
		/* status */
	    u_int       _fs_p2a_full:1;
	    u_int       _fs_p2a_half:1;
	    u_int       _fs_p2a_empty:1;
	    u_int       _fs_a2p_full:1;
	    u_int       _fs_a2p_half:1;
	    u_int       _fs_a2p_empty:1;
	    u_int       _fs_p2a_tczero:1;
	    u_int       _fs_a2p_tczero:1;
		/* control */
	    u_int       _wtc_prio:1;
	    u_int       _wtc_fifo_mng:1;
	    u_int       _wtc_wte:1;
	    u_int       :1;
	    u_int       _rtc_prio:1;
	    u_int       _rtc_fifo_mng:1;
	    u_int       _rtc_rte:1;
	    u_int       :1;
	    u_int       _nv_bus:8;
	    u_int       _reset_ao:1;
	    u_int       _reset_p2a:1;
	    u_int       _reset_a2p:1;
	    u_int       _reset_mb:1;
	    u_int       :1;
	    u_int       _nv_ctl:3;
	} _s;
	__u32       reg;
    } _mcsr;
#define p2afull         _s._fs_p2a_full
#define p2ahalf         _s._fs_p2a_half
#define p2aempty        _s._fs_p2a_empty
#define a2pfull         _s._fs_a2p_full
#define a2phalf         _s._fs_a2p_half
#define a2pempty        _s._fs_a2p_empty
#define p2azero         _s._fs_p2a_tczero
#define a2pzero         _s._fs_a2p_tczero
 
#define wtc_w_prio      _s._wtc_prio
#define wtc_enable      _s._wtc_wte
#define rtc_r_prio      _s._rtc_prio
#define rtc_enable      _s._rtc_rte
#define ao_reset        _s._reset_ao
#define p2a_reset       _s._reset_p2a
#define a2p_reset       _s._reset_a2p
#define mb_reset        _s._reset_mb
#define wtc             _t._wtc
#define rtc             _t._rtc
#define nv_ctl          _t._nv_ctl
#define nv_bus          _t._nv_bus
#define reset           _t._reset
#define status          _t._status
#endif
