/* version to talk to amcc using io ports */
/* you might want to read both /usr/doc/HOWTO/mini/IO-Port-Programming
   and /usr/include/asm/io.h there are other ways to do this */
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>    
#include <time.h>
#include <math.h>                                                               
#include <signal.h>
#include <sys/io.h> /* for io ports and iopl() */
#include <asm/io.h> /* for io ports and iopl() */
#include "amcc.h"

#define AMCC 10e8
#define BOARD 4750
#define SHIFT 14
void print_registers(unsigned long ioaddr);
int getMemconfig();
void get_register(unsigned long ioaddr, unsigned long *target);
void set_register(unsigned long ioaddr, unsigned long target);
void get_pci_info(unsigned long *pcidata);
int writehtonl(int fd, unsigned long *src, unsigned int length);
void gotit(int val);
void disp_mcsr(unsigned int mcsr);
     
int done = 0;

int main(int argc, char **argv) {
  char msg[255], filename[255];
  char writeFlag=0, c;
  FILE *fp;
  int ret, i, cycles=0;
  unsigned long pcidata[5];
  unsigned long bar0, ioaddr;
  unsigned long intcsr, mcsr, mrar, mrtc, mwar, mwtc;
  unsigned long old, bufferSize;
  int fd, filefd;
  char *in_buffer, *out_buffer;
  unsigned long OUT_BUFFER, IN_BUFFER, content=0, *in_buffer_l,
                *out_buffer_l;
  int mem;
  struct timespec tspec, tspec2;
  int shift = 10;

  strcpy(filename, "/tmp/notest.dat");   
  i=0;
  while(i<argc) {
    if(strcasecmp(argv[i],"-b")==0) {
      i++;
      if(argv[i][1]=='x')
        sscanf(argv[i],"%x",&content);
      else
        sscanf(argv[i],"%d",&content);
      i++;                           
    } else if(strcasecmp(argv[i],"-s")==0) {
      i++;
      if(argv[i][1]=='x')
        sscanf(argv[i],"%x",&shift);
      else
        sscanf(argv[i],"%d",&shift);
      i++;                           
    } else if (strcasecmp(argv[i],"-f")==0) {
      strcpy(filename, argv[++i]);
      writeFlag = 1;
      i++;     
    } else {
      i++;
    }
  }

  bufferSize = 1<<shift;

  signal(SIGUSR1, gotit);                                                     

  get_pci_info(pcidata);

  ret=iopl(3); /* grant us access to all the io ports */
  if(ret)
  {
  	printf("Unable to gain access to I/O ports\n");
	return(-1);
  }

  mem = getMemconfig();

  OUT_BUFFER = mem<<20;
  IN_BUFFER = (mem+1)<<20; 

  printf("Content: %lx\n",content);
  fd = open("/dev/mem",O_RDWR);
  if (fd<0) {
    printf("Unable to open BUFFER_ADDR with mmap.\n");
    exit(-1);
  } else {
    out_buffer = (unsigned char*) mmap(0, 0x1000, PROT_READ|PROT_WRITE,
      MAP_SHARED, fd, OUT_BUFFER);
    out_buffer_l = (unsigned long *) out_buffer;
    out_buffer_l[0] = content;
  }
  munmap(out_buffer,0x1000);
  close(fd);

  bar0 = pcidata[3];  /* AMCC Registers */
  printf("bar0: %x\n",bar0);

  if(!(bar0 & 1)) /*if bar0 ends with 0 - memory mapped */
  {
    printf("This version is for I/O mapped devices\n");
    return(-1);
  } /* else its I/O mapped */
	
  printf("Preparing for I/O mapped device\n");
  ioaddr = bar0 & 0xfffc;  /* lose the I/O and reserved bits */
  printf("ioaddr: %x\n",ioaddr);

  intcsr = 0; 
  get_register(ioaddr + AMCC_OP_REG_MCSR, &mcsr);
  mcsr = mcsr &( ~(0xFF));
  printf("in_buffer: %lx\n",IN_BUFFER);  
  printf("out_buffer: %lx\n",OUT_BUFFER);  

  set_register(ioaddr + AMCC_OP_REG_INTCSR, intcsr);

  // RESET FIFO's
  mcsr = (1<<25)  // reset pci to add-on
       | (1<<26); // reset add-on to pci 
  set_register(ioaddr + AMCC_OP_REG_MCSR, mcsr);
  // Set address for reading from the computer to the card
  set_register(ioaddr + AMCC_OP_REG_MRAR, OUT_BUFFER);
  // Set the number of bytes to read from the computer into the card
  set_register(ioaddr + AMCC_OP_REG_MRTC, (1<<2));
  // Set the address for writing from the card into memory
  set_register(ioaddr + AMCC_OP_REG_MWAR, IN_BUFFER);
  // Set the number of bytes to write from the card into memory
  set_register(ioaddr + AMCC_OP_REG_MWTC, bufferSize);
  printf("Finished setup of the cards. \n");
  print_registers(ioaddr);  
  //printf ("Push enter to continue. \n");
  //c = getchar();
  //mcsr = (1<<10) ;  // start add-on to PCI
  //set_register(ioaddr + AMCC_OP_REG_MCSR, mcsr);
  mcsr = (1<<14)   // start pci to add-on
       | (1<<10)   // enable add-on to PCI
       | (1<<9)    // request PCI bus when 4+ are full 
       | (1<<8)    // write over read priority 
       ;
  set_register(ioaddr + AMCC_OP_REG_MCSR, mcsr);
  print_registers(ioaddr);  
  done = 0; old = 0;
  if (content == 0) done = 1;
  do {
    get_register(ioaddr + AMCC_OP_REG_MWTC, &mwtc);
    printf("start of do mwtc: %lx\n", mwtc);
    if (mwtc==0) {
     /*
     set_register(ioaddr + AMCC_OP_REG_MWAR, IN_BUFFER);
     set_register(ioaddr + AMCC_OP_REG_MWTC, bufferSize);
     //mcsr = (1<<14) | (1<<10); 
     mcsr = (1<<14)   // start pci to add-on
       | (1<<10)   // enable add-on to PCI
       | (1<<9)    // request PCI bus when 4+ are full 
       | (1<<8)    // write over read priority 
       ;
     */
     //set_register(ioaddr + AMCC_OP_REG_MCSR, mcsr);
     done=1;
     /*
     get_register(ioaddr + AMCC_OP_REG_MCSR, &mcsr);
     printf("reseting: %d\tmcsr:%x\n",cycles,mcsr);
     */
     cycles++;
     old = 0;
    } else {
      //for(i=0;i<(1<<3);i++){;}
      old = mwtc;
    }
  } while(!done);

  mcsr = 0 ;  
  set_register(ioaddr + AMCC_OP_REG_MCSR, mcsr);
  if (writeFlag&&(done==1)) {
    fd = open("/dev/mem",O_RDWR);
    if (fd<0) {
      printf("Unable to open BUFFER_ADDR with mmap.\n");
      exit(-1);
    } else {
      in_buffer = (unsigned char*) mmap(0, bufferSize, PROT_READ|PROT_WRITE,
        MAP_SHARED, fd, IN_BUFFER);
      in_buffer_l = (unsigned long *) in_buffer;
    }
    printf("Writing to a file.\n");
    filefd = open(filename,O_RDWR|O_CREAT|O_TRUNC|O_NONBLOCK, S_IRWXU);
    writehtonl(filefd, in_buffer_l, bufferSize);
    close(filefd);
    munmap(in_buffer,bufferSize);
    close(fd); 
  } else {
    printf("%d ", cycles*(bufferSize));
    printf("done: %d\n",done);
    printf("****************** Error **********************\n");
  }
  
}

void gotit(int val) {
  printf("The DAQ program got the signal: %d\n",val);
  done = 1;
}  

