/* version to talk to amcc using io ports */
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>    
#include <sys/io.h> /* for io ports and iopl() */
#include <asm/io.h> /* for io ports and iopl() */
#include "amcc.h"

#define AMCC 10e8
#define BOARD 4750

void print_registers(unsigned long ioaddr);
int getMemconfig();
void get_register(unsigned long ioaddr, unsigned long *target);
void set_register(unsigned long ioaddr, unsigned long target);
 
int main(int argc, char **argv) {
  char msg[255];
  int ret, pcidata[5],i;
  FILE *fp;
  unsigned long bar0,ioaddr;
  unsigned long intcsr, mcsr, mrar, mrtc;
  int fd;
  char *buffer;
  unsigned long BUFFER_ADDR, content=0, *buffer_l;
  int mem;
 
  i=0;
  while(i<argc) {
    if(strcasecmp(argv[i],"-b")==0) {
      i++;
      if(argv[i][1]=='x')
        sscanf(argv[i],"%x",&content);
      else
        sscanf(argv[i],"%d",&content);
      i++;                           
    } else {
      i++;
    }
  }
  ret=iopl(3); /* grant us access to all the io ports */
  if(ret)
  {
  	printf("Unable to gain access to I/O ports\n");
	return(-1);
  }

  ret = system("rm /tmp/pci.dat");
  ret = system("grep 10e84750 /proc/bus/pci/devices > /tmp/pci.dat");
  /* printf("ret: %d\n",ret); */
  fp = fopen("/tmp/pci.dat","r");
  for (i=0;i<4;i++) {
    fscanf(fp,"%x",pcidata+i);
    //printf("%x\n",*(pcidata+i));
  }
  fclose(fp);
 
  mem = getMemconfig();
  BUFFER_ADDR = mem<<20;
  fd = open("/dev/mem",O_RDWR);
  if (fd<0) {
    printf("Unable to open BUFFER_ADDR with mmap.\n");
    exit(-1);
  } else {
    buffer = (unsigned char*) mmap(0, 0x1000, PROT_READ|PROT_WRITE,
      MAP_SHARED, fd, BUFFER_ADDR);
    buffer_l = (unsigned long *) buffer;
    buffer_l[0] = content;
    printf("Content: %lx\n",content);
  }

  bar0 = pcidata[3];  /* AMCC Registers */
  printf("bar0: %x\n",bar0);

  if(!(bar0 & 1)) /*if bar0 ends with 0 - memory mapped */
  {
  	printf("This version is for I/O mapped devices\n");
	return(-1);
  }
  /* else its I/O mapped */
	
  printf("Preparing for I/O mapped device\n");
  ioaddr = bar0 & 0xfffc;  /* lose the I/O and reserved bits */
  printf("ioaddr: %x\n",ioaddr);

/* you might want to read both /usr/doc/HOWTO/mini/IO-Port-Programming
   and /usr/include/asm/io.h there are other ways to do this */
     
  intcsr = inl(ioaddr + AMCC_OP_REG_INTCSR); 
  mcsr = inl(ioaddr + AMCC_OP_REG_MCSR);   
  mrar = inl(ioaddr  + AMCC_OP_REG_MRAR);
  mrtc = inl(ioaddr + AMCC_OP_REG_MRTC);

  printf ("intcsr: %lx\n",intcsr); 
  printf ("mcsr: %lx\n",mcsr); 
  printf ("mrar: %lx\n",mrar); 
  printf ("mrtc: %lx\n",mrtc); 
  
  mcsr = mcsr | (1<<25) | (1<<26); // RESET FIFO's
  intcsr = 0x3C000000; // set FIFO advance control no endian conversion
  //intcsr = 0x00000000; // set FIFO advance control no endian conversion

  printf("mcsr: %lx\n",mcsr);
  printf("buffer_addr: %lx\n",BUFFER_ADDR);  

  /*
  outl(ioaddr + AMCC_OP_REG_MCSR, mcsr);
  outl(ioaddr + AMCC_OP_REG_MRAR, BUFFER_ADDR);
  outl(ioaddr + AMCC_OP_REG_MRTC, 1);
  */
  set_register(ioaddr + AMCC_OP_REG_INTCSR, intcsr);
  set_register(ioaddr + AMCC_OP_REG_MCSR, mcsr);
  set_register(ioaddr + AMCC_OP_REG_MRAR, BUFFER_ADDR);
  set_register(ioaddr + AMCC_OP_REG_MRTC, (1<<2));

  print_registers(ioaddr);  
 
  mcsr = mcsr | (1<<14) ;
  //outl(ioaddr + AMCC_OP_REG_MCSR, mcsr);
  set_register(ioaddr + AMCC_OP_REG_MCSR, mcsr);
  print_registers(ioaddr);  

  munmap(buffer,0x1000);
  close(fd); 
}

void print_registers(unsigned long ioaddr) {
  unsigned long intcsr, mcsr, mrar, mrtc;
/*
  intcsr = inl(ioaddr + AMCC_OP_REG_INTCSR); 
  mcsr = inl(ioaddr + AMCC_OP_REG_MCSR);   
  mrar = inl(ioaddr  + AMCC_OP_REG_MRAR);
  mrtc = inl(ioaddr + AMCC_OP_REG_MRTC);
*/
  get_register(ioaddr + AMCC_OP_REG_INTCSR, &intcsr);
  get_register(ioaddr + AMCC_OP_REG_MCSR, &mcsr);
  get_register(ioaddr + AMCC_OP_REG_MRAR, &mrar);
  get_register(ioaddr + AMCC_OP_REG_MRTC, &mrtc);

  printf ("intcsr: %lx\n",intcsr); 
  printf ("mcsr: %lx\n",mcsr); 
  printf ("mrar: %lx\n",mrar); 
  printf ("mrtc: %lx\n",mrtc); 
} 

void get_register(unsigned long ioaddr, unsigned long *target) {
  int i;
  char *temp;

  temp = (char *) target;
  for (i=0;i<4;i++) {
    temp[i] = inb(ioaddr+i);
  }
} 

void set_register(unsigned long ioaddr, unsigned long target) {
  int i;
  unsigned char *temp, temp2;
  /*
  printf("Setting: %lx :",ioaddr);
  temp = (char *) (&target);
  for (i=0;i<4;i++) {
    temp2 = temp[i];
    printf(" %x",temp2);
    outb(ioaddr+i,temp2); // wrong order!!!!
  }
  printf("\n");
  */
  printf("Setting: %lx: %lx\n",ioaddr, target);
  outl(target, ioaddr);  // stupid parameter list order  content,addr
}

int getMemconfig() {
  FILE *fp;
  char msg[255], *ptr, temp[3];
  int mem;
 
  fp = fopen("/proc/cmdline","rb");
  fgets(msg,255,fp);
  fclose(fp);
  ptr = (char *) strstr(msg,"mem");
  sscanf(ptr,"%3s=%d",temp,&mem);
 
  return mem;
}               
